# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimString,
    Segment,
)
from driverng.astm.parser import RecursiveList


class DmsTrackingSegmentH(HeaderSegment):
    sender_name = HprimString((5, 1))
    software_version = HprimString((5, 2))
    processing_id = HprimString(12, default="P")
    version = HprimString(13, default="1")
    date = HprimDate(14)

    infos = HprimForeign("M", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class DmsTrackingSegmentMOLD(Segment):
    source = HprimString((3, 1))
    action = HprimString((3, 2))
    sample_id = HprimString(4)
    date = HprimString((5, 1))
    sample_status_text = HprimString((5, 2))
    sample_status_coding = HprimString((5, 3))
    instrument_code = HprimString(6)
    code = HprimList((7, 1))
    # Description d'une erreur
    # Information sur le tube (Cap color)
    information = HprimString(8)
    user = HprimString(9)
    # Test code for host used by host when has been ordered
    test_code = HprimString(10)
    instrument_aliquoter = HprimString(11)


class DmsTrackingSegmentM_source(Segment):  # Classe pour décrire un event
    source = HprimString(3)


class DmsTrackingSegmentM_sample_id(
    DmsTrackingSegmentM_source
):  # Classe pour décrire le sample_id concerné par l'event
    sample_id = HprimString(4)


class DmsTrackingSegmentM_sample_idEvent(
    DmsTrackingSegmentM_sample_id
):  # Classe pour décrire la date de l'event sample_id
    date = HprimString(5)


class DmsTrackingSegmentM_moduleEvent(
    DmsTrackingSegmentM_sample_idEvent
):  # Class pour décrire le module de l'event sample_id
    id_module = HprimString(6)


class DmsTrackingSegmentM_tubeInfo(
    DmsTrackingSegmentM_moduleEvent
):  # Class pour décrire le tubeInfo de l'event sample_id
    tube_info = HprimString(7)


class DmsTrackingSegmentM_laneInfo(
    DmsTrackingSegmentM_tubeInfo
):  # Class pour décrire le laneInfo de l'event sample_id
    input_lane = HprimString(8)


class DmsTrackingSegmentM_instrumentEvent(
    DmsTrackingSegmentM_sample_idEvent
):  # Class pour décrire l'instrumentId de l'event sample_id
    id_instrument = HprimString(6)


class DmsTrackingSegmentM_ROUT(DmsTrackingSegmentM_instrumentEvent):  # Routing
    tests = HprimList(7)


class DmsTrackingSegmentM_QRY(DmsTrackingSegmentM_instrumentEvent):  # Query
    query_test = HprimList(7)


class DmsTrackingSegmentM_STOR(DmsTrackingSegmentM_moduleEvent):  # Storage
    position = HprimList((7, 1))


class DmsTrackingSegmentM_REQ(Segment):  # Requested
    source = HprimString((3, 1))
    action = HprimString((3, 2))
    sample_id = HprimString(4)
    date = HprimString(5)
    user = HprimString(6)
    id_module = HprimString(7)


class DmsTrackingSegmentM_WARN(DmsTrackingSegmentM_instrumentEvent):  # Warning
    error_code = HprimString(7)
    error_description = HprimString(8)


class DmsTrackingSegmentM_INSTRUMENTEXCEPTION(
    DmsTrackingSegmentM_instrumentEvent
):  # Instrument Exception
    exception_code = HprimString(7)
    exception_description = HprimString(8)
    user = HprimString(9)
    test_code = HprimString(10)


class DmsTrackingSegmentM_WEIGHT(DmsTrackingSegmentM_moduleEvent):  # Weight
    tube_weight = HprimString(7)


class DmsTrackingSegmentM_ALIQINFO(
    DmsTrackingSegmentM_sample_idEvent
):  # Alliquoted Information
    secondary_tube_id = HprimString(6)
    champ_vide = HprimString(7)
    volume = HprimString(8)
    destination = HprimString(9)
    tests = HprimList(10)
    aliquoter_id_module = HprimString(11)


class DmsTrackingSegmentM_VOLUME(DmsTrackingSegmentM_moduleEvent):  # Volume
    serum_volume = HprimString(7)
    serum_level = HprimString(8)


class DmsTrackingSegmentM_rackEvent(
    DmsTrackingSegmentM_source
):  # Class pour décrire le rack et le moduleId de l'event source
    date = HprimString(4)
    rack_position = HprimString(5)
    id_module = HprimString(6)


class DmsTrackingSegmentM_ENABLE_DISABLE(
    DmsTrackingSegmentM_source
):  # Enable / Disable
    id_instrument = HprimString(4)
    test_code = HprimString(5)
    date = HprimString(6)
    user = HprimString(7)


class DmsTrackingSegmentM_PLATEINFO(DmsTrackingSegmentM_source):  # Plate Info
    plate_id = HprimString(4)
    sample_id = HprimString(5)
    vial_info = HprimList(6)
    date = HprimString(7)
    id_instrument = HprimString(8)


class DmsTrackingSegmentM_test(DmsTrackingSegmentM_sample_id):  # Test Status
    test_code = HprimString(5)


class DmsTrackingSegmentM_TSC(DmsTrackingSegmentM_test):  # Test Status
    test_status = HprimList((6, 1))
    date_completion = HprimString(7)
    date = HprimString(8)


class DmsTrackingSegmentM_SSU(
    DmsTrackingSegmentM_sample_id
):  # Sample Status(localisation)
    sample_status = HprimList((5, 1))
    id_instrument = HprimString((6, 1))
    instrument_serial = HprimString((6, 2))
    module = HprimString((6, 3))
    lane = HprimString((6, 4))
    rack = HprimString((6, 5))
    position = HprimString((6, 6))
    date = HprimString(7)


class DmsTrackingSegmentM_ECK(DmsTrackingSegmentM_test):  # Event ACK
    order_code = HprimList((6, 1))
    test_status = HprimList((7, 1))
    date = HprimString(8)


class DmsTrackingSegmentM_SORTINFO(DmsTrackingSegmentM_sample_id):  # Sort Info
    id_module = HprimString(5)
    date = HprimString(6)
    test_list = HprimList(7)


class DmsTrackingSegmentL(Segment):
    terminaison_code = HprimString(3)


SEGMENTM_SOURCE_CLASS_MAP = {
    "CKIN": DmsTrackingSegmentM_laneInfo,
    "RELOAD": DmsTrackingSegmentM_laneInfo,
    "REID": DmsTrackingSegmentM_tubeInfo,
    "EXSTO": DmsTrackingSegmentM_tubeInfo,
    "CKALQ": DmsTrackingSegmentM_tubeInfo,
    "VTMCKIN": DmsTrackingSegmentM_tubeInfo,
    "ROUT": DmsTrackingSegmentM_ROUT,
    "SAMP": DmsTrackingSegmentM_instrumentEvent,
    "QRY": DmsTrackingSegmentM_QRY,
    "STOR": DmsTrackingSegmentM_STOR,
    "REMV": DmsTrackingSegmentM_moduleEvent,
    "CENT": DmsTrackingSegmentM_moduleEvent,
    "ALIQ": DmsTrackingSegmentM_moduleEvent,
    "SEAL": DmsTrackingSegmentM_moduleEvent,
    "UNSE": DmsTrackingSegmentM_moduleEvent,
    "DECA": DmsTrackingSegmentM_moduleEvent,
    "RECA": DmsTrackingSegmentM_moduleEvent,
    "REQ": DmsTrackingSegmentM_REQ,
    "WARN": DmsTrackingSegmentM_WARN,
    "DISP": DmsTrackingSegmentM_moduleEvent,
    "SHAKEN": DmsTrackingSegmentM_moduleEvent,
    "TRANSFERRED": DmsTrackingSegmentM_moduleEvent,
    "INSTRUMENTEXCEPTION": DmsTrackingSegmentM_INSTRUMENTEXCEPTION,
    "WEIGHT": DmsTrackingSegmentM_WEIGHT,
    "ALIQINFO": DmsTrackingSegmentM_ALIQINFO,
    "PRIO": DmsTrackingSegmentM_sample_idEvent,
    "VOLUME": DmsTrackingSegmentM_VOLUME,
    "DISPOSED": DmsTrackingSegmentM_rackEvent,
    "ENABLE": DmsTrackingSegmentM_ENABLE_DISABLE,
    "DISABLE": DmsTrackingSegmentM_ENABLE_DISABLE,
    "PLATEINFO": DmsTrackingSegmentM_PLATEINFO,
    "REMOVED": DmsTrackingSegmentM_rackEvent,
    "TSC": DmsTrackingSegmentM_TSC,
    "SSU": DmsTrackingSegmentM_SSU,
    "ECK": DmsTrackingSegmentM_ECK,
    "SORTINFO": DmsTrackingSegmentM_SORTINFO,
}


def get_segment_source(source):
    return SEGMENTM_SOURCE_CLASS_MAP.get(source, DmsTrackingSegmentMOLD)()


def get_commentaire(seg_m, sample):
    commentaires = []

    if seg_m.source == "STOR":
        rack = "-".join(seg_m.position)
        sample.rack = "{}-{}".format(seg_m.id_module, rack)
        # Stockage du rack provenant du module IOM (Sérothéqueur) en
        # indiquant le nom du module
        if seg_m.id_module.startswith("IOM"):
            commentaires.append(sample.rack)
        else:
            commentaires.append(rack)
    elif seg_m.source == "QRY":
        # Liste des tests demandés au DMS
        commentaires.extend(seg_m.query_test)
    elif seg_m.source == "INSTRUMENTEXCEPTION" and seg_m.exception_description:
        # Description error exception ou de warning
        comments = "{} {}".format(seg_m.exception_description, seg_m.test_code)
        commentaires.append(comments)
    elif seg_m.source == "WARN" and seg_m.error_description:
        commentaires.append(seg_m.error_description)
    elif seg_m.source in ("CKIN", "RELOAD") and seg_m.input_lane:
        commentaires.append(seg_m.input_lane)
    elif seg_m.source in ("ALIQINFO") and seg_m.tests:
        if seg_m.secondary_tube_id:
            comment = "{}_{}".format(seg_m.secondary_tube_id, "-".join(seg_m.tests))
        else:
            comment = "{}".format("-".join(seg_m.tests))
        commentaires.append(comment)
    elif seg_m.source in ("VOLUME") and seg_m.serum_level:
        commentaires.append(seg_m.serum_level)
    elif seg_m.source in ("PLATEINFO") and seg_m.plate_id:
        commentaires.append(seg_m.plate_id)
    elif seg_m.source in ("REQ") and seg_m.action:
        commentaires.append(seg_m.action)
    elif seg_m.source in ("DISPOSED", "REMOVED") and seg_m.rack_position:
        commentaires.append(seg_m.rack_position)
    elif seg_m.source in ("ENABLE", "DISABLE") and seg_m.test_code:
        commentaires.append(seg_m.test_code)
    elif seg_m.source in ("SSU") and seg_m.sample_status:
        commentaires.append("-".join(seg_m.sample_status))
    return commentaires


def get_flags(seg_m):
    flags = []

    if (
        seg_m.source in ("CKIN", "RELOAD", "REID", "EXSTO", "CKALQ", "VTMCKIN")
        and seg_m.tube_info
    ):
        flags.append(seg_m.tube_info)
    elif seg_m.source in ("ROUT") and seg_m.tests:
        flags.extend(seg_m.tests)
    elif seg_m.source in ("REQ") and seg_m.id_module:
        flags.append(seg_m.id_module)
    elif seg_m.source in ("WARN") and seg_m.error_code:
        flags.append(seg_m.error_code)
    elif seg_m.source in ("INSTRUMENTEXCEPTION") and seg_m.exception_code:
        flags.append(seg_m.exception_code)
    elif seg_m.source in ("WEIGHT") and seg_m.tube_weight:
        flags.append(seg_m.tube_weight)
    elif seg_m.source in ("ALIQINFO") and seg_m.volume:
        flags.append(seg_m.volume)
    elif seg_m.source in ("VOLUME") and seg_m.serum_volume:
        flags.append(seg_m.serum_volume)
    elif seg_m.source in ("ECK") and seg_m.test_status:
        flags.append(seg_m.test_status)
    elif seg_m.source in ("SORTINFO") and seg_m.test_list:
        flags.extend(seg_m.test_list)
    return flags


def get_date(seg_m):
    try:
        date = datetime.datetime.strptime(seg_m.date, "%Y%m%d%H%M%S")
    except ValueError:
        date = None
    return date


def get_instrument(nom_chaine_dms, seg_m):
    instrument = ""
    if seg_m.source == "ALIQINFO":
        if nom_chaine_dms:
            instrument = "{}_{}".format(nom_chaine_dms, seg_m.aliquoter_id_module)
        else:
            instrument = seg_m.aliquoter_id_module
    elif (
        seg_m.source
        in (
            "CKIN",
            "RELOAD",
            "REID",
            "EXSTO",
            "CKALQ",
            "VTMCKIN",
            "STOR",
            "REMV",
            "CENT",
            "ALIQ",
            "SEAL",
            "UNSE",
            "DECA",
            "RECA",
            "REQ",
            "DISP",
            "SHAKEN",
            "TRANSFERRED",
            "WEIGHT",
            "VOLUME",
            "DISPOSED",
            "REMOVED",
            "SORTINFO",
        )
        and seg_m.id_module
    ):
        if nom_chaine_dms:
            instrument = "{}_{}".format(nom_chaine_dms, seg_m.id_module)
        else:
            instrument = seg_m.id_module
    elif (
        seg_m.source
        in (
            "ROUT",
            "SAMP",
            "QRY",
            "WARN",
            "INSTRUMENTEXCEPTION",
            "ENABLE",
            "DISABLE",
            "PLATEINFO",
            "SSU",
        )
        and seg_m.id_instrument
    ):
        if nom_chaine_dms:
            instrument = "{}_{}".format(nom_chaine_dms, seg_m.id_instrument)
        else:
            instrument = seg_m.id_instrument
    return instrument


def get_date_plateau(seg_m):
    date_plateau = None
    if seg_m.source == "CKIN":
        date_plateau = datetime.datetime.strptime(seg_m.date, "%Y%m%d%H%M%S")
    return date_plateau


def get_operator(seg_m):
    operator = None
    if (
        seg_m.source in ("REQ", "INSTRUMENTEXCEPTION", "ENABLE", "DISABLE")
        and seg_m.user
    ):
        operator = seg_m.user
    return operator


def get_sample_id(seg_m):
    try:
        sample_id = seg_m.sample_id
    except (ValueError, AttributeError):
        sample_id = ""
    return sample_id


class Driver(AstmMessageV2):
    SPEC = {"code": "dms_tracking", "label": "Dms Tracking Flexlab"}
    SEGMENT_CLASS_MAP = {
        "H": DmsTrackingSegmentH,
        "M": DmsTrackingSegmentM_ENABLE_DISABLE,
        "L": DmsTrackingSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    OPTIONS = [
        AstmMessageV2.Option(
            "nom_chaine_dms",
            label="Nom de la chaîne DMS dans le cas de plusieurs chaines connectées",
            default="DMS",
        )
    ]

    def __init__(self, nom_chaine_dms="DMS", **kwargs):
        self.nom_chaine_dms = nom_chaine_dms
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = DmsTrackingSegmentH(type_segment="H", **kwargs)
        message.end = DmsTrackingSegmentL(type_segment="L")
        return message

    def parse_segment(self, line, separators):
        segment = super(Driver, self).parse_segment(line, separators)
        if segment.type_segment != "M":
            return segment

        seg_m = get_segment_source(segment.source)

        splitted = RecursiveList.split(line, separators)
        seg_m.populate(splitted)
        return seg_m

    def encode_worklist(self, worklist, context):
        message = self.create(date=context["now"])
        for sample in worklist.samples:
            for order in sample.orders:
                if order.action == "del":
                    source = "DISABLE"
                else:
                    source = "ENABLE"
                seg_m = message.infos.add(source=source, id_instrument=sample.sample_id)
                seg_m.set_raw("test_code", order.code)
        if message.infos:
            return message

    def decode_message(self, message):
        worklist = Worklist(context="result")
        for seg_m in message.infos:
            # Ne pas traiter les évènements de traçabilité de source TSC et ECK
            # non pertinent
            if seg_m.source in ("TSC", "ECK"):
                continue
            sample_id = get_sample_id(seg_m)
            if not sample_id:
                continue
            sample = worklist.get_or_create(sample_id=sample_id, type="patient")
            commentaires = get_commentaire(seg_m, sample)
            flags = get_flags(seg_m)
            date = get_date(seg_m)
            instrument = get_instrument(self.nom_chaine_dms, seg_m)
            date_plateau = get_date_plateau(seg_m)
            operator = get_operator(seg_m)
            sample.results.add(
                code="TRACKING",
                value=seg_m.source,
                flags=flags,
                comments=commentaires,
                date=date,
                instrument=instrument,
                date_plateau=date_plateau,
                operator=operator,
            )
        return worklist
